import os
import functools

import pandas as pd
import numpy as np

from nltk import word_tokenize
from gensim import corpora
from gensim import models
from gensim import similarities
from collections import defaultdict, Counter

from pruvo.preprocess import pre_process

AVAILABLE_MODELS = {
    'tfidf': {
        'model': models.TfidfModel,
        'train_vector_fn': None,
        'sim_vector_fn': None,
        'args': [],
        'kwargs': {
            'normalize': True,
        },
        'return_method': None,
    },
    'lsi': {
        'model': models.LsiModel,
        'train_vector_fn': '_get_tfidf_vectors',
        'sim_vector_fn': None,
        'args': [],
        'kwargs': {
            'id2word': None,
            'num_topics': 300,
        },
        'return_method': None,
    },
    'rp': {
        'model': models.RpModel,
        'train_vector_fn': '_get_tfidf_vectors',
        'sim_vector_fn': None,
        'args': [],
        'kwargs': {
            'num_topics': 500,
        },
        'return_method': None,
    },
    'lda': {
        'model': models.LdaModel,
        'train_vector_fn': None,
        'sim_vector_fn': None,
        'args': [],
        'kwargs': {
            'id2word': None,
            'num_topics': 100,
        },
        'return_method': None,
    },
    'hdp**experimental**': {
        'model': models.HdpModel,
        'train_vector_fn': None,
        'sim_vector_fn': None,
        'args': [],
        'kwargs': {
            'id2word': None,
        },
        'return_method': None,
    },
    'word2vec': {
        'model': models.Word2Vec,
        'train_vector_fn': '_get_w2v_words',
        'sim_vector_fn': '_get_sim_sif_vectors',
        'args': [],
        'kwargs': {},
        'return_method': 'wv',
    },
}
MODELS_PATH = 'data/models'
ALPHA = 0.001


class Corpus():
    remove_only_once = True
    _text_corpus = []
    _processed_corpus = []
    _dictionary = None
    _bow_corpus = []
    _trained_model = None
    _similarity_index = None
    _frequency = None
    _tfidf_corpus = []
    _model = ''
    _train_vectors = []
    _embedding_size = 0
    _total_words = 0
    
    def add(self, documents=[], pre_processed=False):
        """Add documents to the Corpus

        Parameters
        ----------
        documents : list
            Text sentences to include in the corpus
        pre_processed : boolean
            Indicates whether the documents have been previously preprocessed (using the pre_process function in pruvo.preprocess)
        
        Returns
        -------
        None
        
        """
        
        if not pre_processed:
            pp_list = []
            for doc in documents:
                text = pre_process(doc)
                if text not in pp_list:
                    pp_list.append(text)
        else:
            pp_list = documents
            
        self._text_corpus = self._unique(self._text_corpus + pp_list)
        self._process(pre_processed=pre_processed)
        
    def get_corpus(self, as_df=True):
        """Obtain all documents in Corpus

        Parameters
        ----------
        as_df : boolean
            Return documents as a pandas DataFrame
        
        Returns
        -------
        list | pandas.DataFrame
            All documents in Corpus
        
        """
        if len(self._text_corpus) == 0:
            raise Exception('No corpus available. Run add method')
        return pd.DataFrame(self._text_corpus, columns=['document']) if as_df else self._text_corpus
    
    def get_processed(self, as_df=True):
        """Obtain the list of tokens (words) for all documents in Corpus

        Parameters
        ----------
        as_df : boolean
            Return processed documents as a pandas DataFrame
        
        Returns
        -------
        list | pandas.DataFrame
            List of tokens for all documents in Corpus
        
        """
        if len(self._processed_corpus) == 0:
            raise Exception('No corpus available. Run add method')
        return pd.DataFrame(np.array(self._processed_corpus, dtype=object), columns=['word_tokens']) if as_df else self._processed_corpus
    
    def get_dictionary(self):
        """Obtain the gensim Dictionary for the Corpus
        
        Parameters
        ----------
        None
        
        Returns
        -------
        `gensim.corpora.dictionary.Dictionary`
            gensim Dictionary for the Corpus
        
        """
        if len(self._text_corpus) == 0:
            raise Exception('No corpus available. Run add method')
        return self._dictionary
    
    def get_vectors(self, as_df=False):
        """Obtain the list of basic vectors for all documents in Corpus
        
        Parameters
        ----------
        as_df : boolean
            Return list of vectors as a pandas DataFrame
        
        Returns
        -------
        list | pandas.DataFrame
            List of basic vectors for all documents in Corpus.
            A basic vector is a tuple, where first number is the indice of the word in dictionary
            and the second number is the frequency of the word on document
        
        """
        if len(self._bow_corpus) == 0:
            raise Exception('No dictionary available. Run add method')
        return pd.DataFrame(np.array(self._bow_corpus, dtype=object), columns=['vector']) if as_df else self._bow_corpus
    
    def tokenize(self, document, pre_processed=False):
        """Tokenize (keep only the important words) a document
        
        Parameters
        ----------
        document : string
            A text sentence
        pre_processed : boolean
            Indicates whether the document have been previously preprocessed (using the pre_process function in pruvo.preprocess)
        
        Returns
        -------
        list
            List of the important words in document
        
        """
        if not pre_processed:
            document = pre_process(document)
        return word_tokenize(document)
    
    def vectorize(self, document, pre_processed=False):
        """Vectorize (important word frequencies) a document
        
        Parameters
        ----------
        document : string
            A text sentence
        pre_processed : boolean
            Indicates whether the document have been previously preprocessed (using the pre_process function in pruvo.preprocess)
        
        Returns
        -------
        list
            List of the basic vector in document
            A basic vector is a tuple, where first number is the indice of the word in dictionary
            and the second number is the frequency of the word on document
        
        """
        if len(self._bow_corpus) == 0:
            raise Exception('No dictionary available. Run add method')
        return self._dictionary.doc2bow(self.tokenize(document, pre_processed))
    
    def get_frequencies(self, counter=False):
        """Get frequency of the words on Corpus
        
        Parameters
        ----------
        counter : boolean
            Return frequencies as a collections.Counter
        
        Returns
        -------
        list | collections.Counter
            List of tuples where first value is the word
            and the second number is the count of this word in Corpus
        
        """
        if counter:
            return Counter(self._frequency)
        else:
            return sorted(self._frequency.items(), key=lambda word: word[1], reverse=True)

    def train(self, model, calc_sim=True, *args, **kwargs):
        """Train a NLP model against the Corpus
        
        Parameters
        ----------
        model : string
            The name of the model
        calc_sim : boolean
            Calculate the similarity matrix of the Corpus
        
        Returns
        -------
        gensim.models.base_any2vec.BaseAny2VecModel
            The trained model used to get the word/sentence vectors
        
        """
        
        if model not in AVAILABLE_MODELS.keys():
            raise Exception('Model not available')
            
        self._model = model            
        if not os.path.exists(MODELS_PATH):
            os.makedirs(MODELS_PATH)
            
        model_config = AVAILABLE_MODELS[model]
        
        if model_config['train_vector_fn'] is not None:
            vector_method = getattr(self, model_config['train_vector_fn'], None)
            self._train_vectors = vector_method()
        else:
            self._train_vectors = self.get_vectors()
            
        if 'id2word' in model_config['kwargs'] and model_config['kwargs']['id2word'] is None:
            model_config['kwargs']['id2word'] = self.get_dictionary()
            
        self._trained_model = model_config['model'](self._train_vectors, *model_config['args'], **dict(model_config['kwargs'], **kwargs))

        if model_config['return_method'] is not None:
            model_method = getattr(self._trained_model, model_config['return_method'], None)
            self._trained_model = model_method
            
        self._embedding_size = self._trained_model[0].size
            
        if model == 'tfidf':
            self._tfidf_corpus = self._trained_model[self._bow_corpus]
            
        if calc_sim:
            
            if model_config['sim_vector_fn'] is not None:
                vector_method = getattr(self, model_config['sim_vector_fn'], None)
                sim_vectors = vector_method()
            else:
                sim_vectors = self._get_sim_vectors()
                
            self._similarity_index = similarities.Similarity(MODELS_PATH + f'/{model}', sim_vectors, num_features=len(self._dictionary), shardsize=50000)
        
        return self._trained_model
        
    def get_similars(self, document):
        """Get similar documents
        
        Parameters
        ----------
        document : string
            A text sentence
        
        Returns
        -------
        pandas.DataFrame
            DataFrame with all documents in Corpus and the similarity with the informed document
        
        """
        
        if self._model == 'word2vec':
            doc_vector = self._sif_vector(document)
        else:
            doc_vector = self._trained_model[self.vectorize(document)]
            
        sims = self._similarity_index[doc_vector]

        df_sims = pd.DataFrame(list(enumerate(sims)), columns=['index', 'similarity']).drop(columns=['index'])
        similars = df_sims[df_sims.similarity > 0].sort_values('similarity', ascending=False)
        similars = similars.join(self.get_corpus())
        
        return similars
    
    def _process(self, pre_processed=False):
        self._processed_corpus = []
        for document in self._text_corpus:
            self._processed_corpus.append(self.tokenize(document, pre_processed=pre_processed))
            
        self._frequency = defaultdict(int)
        for doc in self._processed_corpus:
            for token in doc:
                self._frequency[token] += 1

        self._total_words = functools.reduce(lambda a, b: a+b, [self._frequency[word] for word in self._frequency.keys()])
            
        if self.remove_only_once:
            self._processed_corpus = [
                [token for token in doc if self._frequency[token] > 1]
                for doc in self._processed_corpus
            ]            
        
        self._dictionary = corpora.Dictionary(self._processed_corpus)
        self._bow_corpus = [self._dictionary.doc2bow(text) for text in self._processed_corpus]
    
    def _unique(self, list_aux):
        unique = sorted(list(set(list_aux)))
        if len(unique) > 0 and unique[0] == '':
            unique.remove('')
        return unique

    def _sif_vector(self, document, as_tuple=True):

        if self._trained_model is None:
            raise Exception('No trained model available. Run train method')

        tokens = [token for token in self.tokenize(document) if token in self._trained_model.index_to_key]

        doc_vector = np.zeros(self._embedding_size)
        num_words = len(tokens)
        if num_words > 0:
            for word in tokens:
                alpha_word = ALPHA / (ALPHA + (self._frequency[word] / self._total_words))
                doc_vector = np.add(doc_vector, np.multiply(alpha_word, self._trained_model[word]))
            doc_vector = np.divide(doc_vector, num_words)

        if as_tuple:
            tuple_vector = []
            for index, v in enumerate(doc_vector):
                tuple_vector.append((index, v))
            doc_vector = tuple_vector

        return doc_vector

    def _get_tfidf_vectors(self):
        if len(self._tfidf_corpus) == 0:
            self.train(model='tfidf', calc_sim=False)
        if len(self._tfidf_corpus) == 0:
            raise Exception('No TF-IDF model available. Run train(model="tf-idf") method')
        return self._tfidf_corpus
    
    def _get_w2v_words(self):
        return self.get_processed(as_df=False)
    
    def _get_sim_vectors(self):
        return self._trained_model[self._train_vectors]

    def _get_sim_sif_vectors(self):
        vectors = []
        for doc in self._text_corpus:
            vectors.append(self._sif_vector(doc))
        return vectors 
